from pathlib import Path

import pytest

from garbanzo.helpers import get_relative_account_dir, replace_dict_keys, string_is_date


def test_replace_dict_keys():
    d = {'a': 1, 'b': 2}
    assert replace_dict_keys(d, {'a': 'A', 'c': 'C'}) == {'A': 1, 'b': 2}
    with pytest.raises(ValueError, match='one-to-one'):
        _ = replace_dict_keys(d, {'a': 'A', 'c': 'A'})

@pytest.mark.parametrize('account', [
    '',
    'Assets',
    'Assets:Bank',
    'Assets:Bank:RothIRA',
])
def test_get_relative_account_dir(account):
    path = Path('/'.join(account.split(':')))
    assert get_relative_account_dir(account) == path

@pytest.mark.parametrize(['string', 'is_date'], [
    ('2025-01-01', True),
    ('2025-1-01', True),
    ('2025-1-1', True),
    ('', False),
    ('01/01/2025', False),
])
def test_string_is_date(string, is_date):
    assert string_is_date(string) == is_date



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass, field
from functools import reduce
import json
import operator
from pathlib import Path
import re
from typing import Any, Callable, Generic, Optional, Self, TypeVar, cast

from beancount.core.data import Account, Directive, Posting, Transaction
from fancy_dataclass import JSONDataclass

from garbanzo.helpers import replace_dict_keys
from garbanzo.ledger import BeancountLedger, LedgerTransform


S = TypeVar('S')
T = TypeVar('T', bound=Directive)

FilterFunc = Callable[[T], Optional[T]]
PredFunc = Callable[[T], bool]


###########
# HELPERS #
###########


class _Unset:
    """Sentinel value indicating to set a field's value to None."""


UNSET = _Unset()


def _none_if_unset(obj: Optional[S] | _Unset) -> Optional[S]:
    if obj == UNSET:
        return None
    return cast(Optional[S], obj)


# identity function
identity = lambda obj: obj


##############
# FILTER API #
##############


class BaseEntryFilter(ABC):
    """Filters a list of entries.
    This takes each entry and either filters it out (by returning None) or alters it (by returning a new entry)."""

    @abstractmethod
    def __call__(self, entry: Directive) -> Optional[Directive]:
        """Given an entry, returns None if the entry is to be removed, and a new entry if it is to be modified."""


class EntryFilter(BaseEntryFilter, Generic[T]):
    """Concrete implementation of an entry filter.
    This wraps a user-provided filter function, which takes an entry and returns an (optional) entry."""

    def __init__(self, func: FilterFunc[T]) -> None:
        self.func = func

    def __call__(self, entry: Directive) -> Optional[Directive]:
        """Calls the user-defined function on an entry."""
        assert isinstance(entry, Directive)
        return self.func(cast(T, entry))

    def __rshift__(self, other: Self) -> EntryFilter[T]:
        """Sequential composition of filters.
        filter1 >> filter2 means to apply filter1 first, then pass the output (if not None) to filter2."""

        def _composed(entry: T) -> Optional[T]:
            val = self(entry)
            if val is None:
                return None
            return other(val)

        return EntryFilter(_composed)

    @classmethod
    def chain(cls, filters: Sequence[EntryFilter[T]]) -> EntryFilter[T]:
        """Chains together multiple EntryFilters into a single filter by applying them sequentially."""
        return ChainedEntryFilter(list(filters))


@dataclass
class ChainedEntryFilter(EntryFilter[T]):
    """An EntryFilter consisting of a chain of EntryFilters applied sequentially."""

    filters: list[EntryFilter[T]] = field(default_factory=list)

    def __post_init__(self) -> None:
        if self.filters:
            func: FilterFunc[T] = reduce(operator.rshift, self.filters)
        else:
            func = identity
        super().__init__(func)


class TransactionFilter(EntryFilter[Transaction]):
    """Specialization of EntryFilter for Transactions."""

    def __call__(self, entry: Directive) -> Optional[Directive]:
        """If the entry is a Transaction, call the filter funciton; otherwise leave it alone."""
        if isinstance(entry, Transaction):
            return self.func(entry)
        assert isinstance(entry, Directive)
        return entry


class BaseEntryPredicate(ABC):
    """A predicate (bool-valued function) on entries."""

    @abstractmethod
    def __call__(self, entry: Directive) -> bool:
        """Given an entry, returns True or False indicating whether the entry satisfies the predicate."""


class EntryPredicate(BaseEntryPredicate, Generic[T]):
    """Concrete implementation of an entry predicate.
    This wraps a user-provided predicate function, which takes an entry and returns True or False."""

    pred: Optional[PredFunc[T]] = None

    def __init__(self, pred: PredFunc[T]) -> None:
        self.pred = pred

    def __call__(self, entry: Directive) -> bool:
        """Calls the user-defined predicate on an entry."""
        assert self.pred is not None
        assert isinstance(entry, Directive)
        return self.pred(cast(T, entry))

    def __and__(self, other: EntryPredicate[T]) -> EntryPredicate[T]:
        """Takes the conjunction (logical AND) of two EntryPredicates."""

        def _pred(entry: T) -> bool:
            return self(entry) and other(entry)

        return EntryPredicate(_pred)

    def __or__(self, other: EntryPredicate[T]) -> EntryPredicate[T]:
        """Takes the disjunction (logical OR) of two EntryPredicates."""

        def _pred(entry: T) -> bool:
            return self(entry) or other(entry)

        return EntryPredicate(_pred)

    def __invert__(self) -> EntryPredicate[T]:
        """Takes the complement (logical NOT) of an EntryPredicate."""

        def _pred(entry: T) -> bool:
            return not self(entry)

        return EntryPredicate(_pred)

    def to_entry_filter(self) -> EntryFilter[T]:
        """Converts the EntryPredicate into an EntryFilter.
        This will take an entry and return the same entry if the predicate is satisfied, and None otherwise."""

        def _func(entry: T) -> Optional[T]:
            return entry if self(entry) else None

        return EntryFilter(_func)


class TransactionPredicate(EntryPredicate[Transaction]):
    """Specialization of EntryPredicate for Transactions."""


def filter_transform(filt: BaseEntryFilter | Callable[[BeancountLedger], BaseEntryFilter]) -> LedgerTransform:
    """Wraps a BaseEntryFilter into a LedgerTransform that applies the filter to each entry."""

    def _transform(ledger: BeancountLedger) -> BeancountLedger:
        if isinstance(filt, BaseEntryFilter):
            entry_filter = filt
        else:
            # assume that filt is a function that takes a ledger and returns a BaseEntryFilter
            entry_filter = filt(ledger)
        entries = []
        for entry in ledger.entries:
            entry = entry_filter(entry)
            # TODO: count number of entries?
            if entry is not None:
                entries.append(entry)
        return BeancountLedger(ledger.path, entries, [], ledger.options_map)

    return _transform


############
# MATCHING #
############


def _field_matches_regex(field: str, regex: str, flag: re.RegexFlag = re.IGNORECASE) -> PredFunc[object]:
    r = re.compile(regex, flag)

    def _pred(obj: object) -> bool:
        val = getattr(obj, field)
        return isinstance(val, str) and bool(r.findall(val))

    return _pred


@dataclass
class FieldMatchesRegex(EntryPredicate[T]):
    """Checks if a certain field of an Entry matches a particular regular expression."""

    field: str
    regex: str
    flag: re.RegexFlag = re.IGNORECASE

    def __post_init__(self) -> None:
        super().__init__(_field_matches_regex(self.field, self.regex, self.flag))


@dataclass
class HasAccountPrefix(TransactionPredicate):
    """Checks if any posting has an account that starts with a particular prefix."""

    account_prefix: str

    def __post_init__(self) -> None:
        def _pred(txn: Transaction) -> bool:
            return any(posting.account.startswith(self.account_prefix) for posting in txn.postings)

        super().__init__(_pred)


@dataclass
class HasAccountPrefixes(TransactionPredicate):
    """Checks if any posting has an account that starts with each prefix in a list of prefixes."""

    account_prefixes: list[str]

    def __post_init__(self) -> None:
        if not self.account_prefixes:
            raise ValueError('must provide at least one account prefix')
        # take the conjunction of HasAccountPrefix predicate for each prefix
        pred = reduce(operator.and_, map(HasAccountPrefix, self.account_prefixes))
        super().__init__(pred)


##########
# UPDATE #
##########


@dataclass
class AssignField(EntryFilter[T]):
    """Assigns a field to some value."""

    field: str  # name of the field
    value: T  # value to assign

    def __post_init__(self) -> None:
        kwargs = {self.field: self.value}
        super().__init__(lambda entry: entry._replace(**kwargs))


@dataclass
class AssignAccount(TransactionFilter):
    """Given a Transaction, if it has a unique posting whose account begins with account_prefix, will change the account of the posting to the given one."""

    account_prefix: str
    account: str

    def __post_init__(self) -> None:
        assert self.account.startswith(self.account_prefix)
        posting_matches_prefix: PredFunc[Posting] = lambda posting: posting.account.startswith(self.account_prefix)

        def _assign(txn: Transaction) -> Optional[Transaction]:
            valid_postings = list(filter(posting_matches_prefix, txn.postings))
            if len(valid_postings) == 1:
                postings = [posting._replace(account=self.account) if posting_matches_prefix(posting) else posting for posting in txn.postings]
                return txn._replace(postings=postings)
            return txn

        super().__init__(_assign)


@TransactionFilter
def trim_whitespace(txn: Transaction) -> Optional[Transaction]:
    """Remove excess whitespace from narration and payee fields."""
    kwargs = {}
    for key in ['narration', 'payee']:
        val = getattr(txn, key)
        if val is not None:
            kwargs[key] = ' '.join(val.strip().split())
    return txn._replace(**kwargs) if kwargs else txn


################
# JSON FILTERS #
################


class JSONTransactionFilter(TransactionFilter, JSONDataclass, store_type='off'):  # type: ignore[misc]
    """Subclass of TransactionFilter using JSON-specified criteria."""


class JSONTransactionPredicate(TransactionPredicate, JSONDataclass, store_type='off'):  # type: ignore[misc]
    """Subclass of TransactionFilter using JSON-specified criteria."""


@dataclass
class TransactionMatch(JSONTransactionPredicate, suppress_none=True):
    """Class for matching a Transaction based on JSON-specified criteria."""

    narration: Optional[str] = None  # regex to match narration
    payee: Optional[str] = None  # regex to match payee
    account_prefixes: Optional[list[str]] = None  # account prefixes to match

    def __post_init__(self) -> None:
        preds: list[PredFunc[Transaction]] = []
        for fld in ['narration', 'payee']:
            val = getattr(self, fld)
            if val is not None:
                preds.append(FieldMatchesRegex(fld, val))
        if self.account_prefixes is not None:
            preds.append(HasAccountPrefixes(self.account_prefixes))
        if preds:
            pred: PredFunc[Transaction] = reduce(operator.and_, preds)
        else:
            pred = lambda _: True
        super().__init__(pred)


@dataclass
class TransactionUpdate(JSONTransactionFilter, suppress_none=True):
    """Class for updating a Transaction based on JSON-specified criteria."""

    # narration to set
    narration: Optional[_Unset | str] = None
    # payee to set
    payee: Optional[_Unset | str] = None
    # tags to set
    tags: Optional[_Unset | list[str]] = None
    # links to set
    links: Optional[_Unset | list[str]] = None
    # account to set (for a unique posting matching this account type)
    # For example, if the value is 'Expenses:Food:Restaurants', this will set an 'Expense' posting to this specific account, provided there is a unique posting with the 'Expense' prefix.
    account: Optional[str] = None

    def __post_init__(self) -> None:
        filters: list[EntryFilter[Transaction]] = []
        for fld in ['narration', 'payee', 'tags', 'links']:
            # If value is None, leave the field alone.
            # If UNSET, set it to None.
            if (value := getattr(self, fld)) is not None:
                value = _none_if_unset(value)
                filters.append(AssignField(fld, value))
        if self.account is not None:
            account_prefix = self.account.split(':')[0]
            filters.append(AssignAccount(account_prefix, self.account))
        super().__init__(ChainedEntryFilter(filters))

    def to_dict(self, **kwargs: Any) -> dict[str, Any]:  # noqa: D102
        d = super().to_dict(**kwargs)
        # NOTE: we interpret UNSET as JSON null, and None as a missing field
        for fld in ['narration', 'payee']:
            if getattr(self, fld) == UNSET:
                d[fld] = None
        return d

    @classmethod
    def from_dict(cls, d: dict[str, Any], **kwargs: Any) -> Self:  # noqa: D102
        # NOTE: we interpret JSON null as UNSET, and a missing field as None
        d2 = {key: UNSET if (val is None) else val for (key, val) in d.items()}
        return super().from_dict(d2, **kwargs)


@dataclass
class FixTransaction(JSONTransactionFilter):
    """Given a Transaction, if it matches the "match" criteria, applies the "set" criteria.
    The criteria can be specified with a JSON blob with "match" and "set" keys."""

    match_: TransactionMatch
    set: TransactionUpdate

    @staticmethod
    def _fix_dict(d: dict[str, Any]) -> dict[str, Any]:
        return {('match' if (key == 'match_') else key): val for (key, val) in d.items()}

    def to_dict(self, **kwargs: Any) -> dict[str, Any]:  # noqa: D102
        return replace_dict_keys(super().to_dict(**kwargs), {'match_': 'match'})

    @classmethod
    def from_dict(cls, d: dict[str, Any], **kwargs: Any) -> Self:  # noqa: D102
        d = replace_dict_keys(d, {'match': 'match_'})
        return super().from_dict(d, **kwargs)

    def __post_init__(self) -> None:
        def _func(txn: Transaction) -> Optional[Transaction]:
            if self.match_(txn):  # if transaction matches, apply the update
                return self.set(txn)
            # otherwise leave the transaction alone
            return txn

        super().__init__(_func)


def load_json_filters(json_path: Path) -> list[FixTransaction]:
    """Loads a JSON file containing a list of parameter objects that can be converted to FixTransaction objects, and returns the list of such objects."""
    with open(json_path) as f:
        obj_list = json.load(f)
    err = 'JSON file must contain a list of JSON objects'
    assert isinstance(obj_list, list), err
    assert all(isinstance(obj, dict) for obj in obj_list), err
    filters = []
    for obj in obj_list:
        filters.append(FixTransaction.from_dict(obj))
    return filters


#############
# TRANSFERS #
#############


@dataclass
class RenameTransfers(TransactionFilter):
    """Sets the narration of any "transfer" transaction in a consistent format, "TYPE: SRC_NAME -> DEST_NAME".

    A transfer is any transaction consisting of two postings, each of whose account is in account_name_map, which maps from each account to its name.
    TYPE is "Transfer" if the destination account is an Asset, and "Payment" if the destination account is a Liability.
    SRC_NAME is the name associated with the source account.
    DEST_NAME is the name associated with the destination account."""

    account_name_map: dict[Account, str]

    def __post_init__(self) -> None:
        name_map = self.account_name_map

        def _rename_transfer(txn: Transaction) -> Optional[Transaction]:
            postings = txn.postings
            if (len(postings) == 2) and all(posting.account in name_map for posting in postings):
                # first account is the source (the one that was debited)
                if postings[0].units.number < 0:
                    src, dest = postings[0].account, postings[1].account
                else:
                    src, dest = postings[1].account, postings[0].account
                if dest.startswith('Assets'):
                    label = 'Transfer'
                elif dest.startswith('Liabilities'):
                    label = 'Payment'
                else:  # no valid label
                    return txn
                src_name = name_map[src]
                dest_name = name_map[dest]
                narration = f'{label}: {src_name} -> {dest_name}'
                return txn._replace(narration=narration)
            return txn

        super().__init__(_rename_transfer)

    @classmethod
    def from_ledger(cls, ledger: BeancountLedger) -> Self:
        """Constructs a RenameTransfers filter from a BeancountLedger by constructing the account_name_map from the set of all Open directives specifying a "name" metadata field."""
        account_name_map = {}
        for entry in ledger.open_directives:
            if entry.meta and ('name' in entry.meta):
                account = entry.account
                name = entry.meta['name']
                if account.startswith('Assets') or account.startswith('Liabilities'):
                    account_name_map[account] = name
        return cls(account_name_map)


rename_transfers = filter_transform(RenameTransfers.from_ledger)



from collections import defaultdict
from collections.abc import Sequence
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Optional, Self

from beancount import loader
import beancount.core.data
from beancount.core.data import Directive, Open, Transaction
from beancount.parser import options
from beancount.parser.grammar import ParserSyntaxError
from beancount.parser.lexer import LexerError

from garbanzo import logger
from garbanzo.helpers import quote, write_lines


# function to modify a BeancountLedger (immutably)
LedgerTransform = Callable[['BeancountLedger'], 'BeancountLedger']

# option keys not to include when writing out options to ledger file
EXEMPT_OPTION_KEYS = {'dcontext', 'filename', 'include', 'input_hash', 'plugin'}


def has_valid_beancount_extension(path: Path) -> bool:
    """Returns True if the given path has a valid extension for a beancount file."""
    return path.suffix.lower() in ['.beancount', '.bean']


def _apply_file_extension(filename: str, extension: str) -> str:
    """Applies an extension to a filename string."""
    return str(Path(filename).with_suffix(extension))


LEDGER_FATAL_ERR_TYPES = (
    LexerError,
    loader.LoadError,
    ParserSyntaxError,
)


def _ledger_error_is_fatal(error: object) -> bool:
    """Returns True if an error should be considered fatal when parsing a beancount ledger."""
    if isinstance(error, LEDGER_FATAL_ERR_TYPES):
        return True
    return False


class LedgerError(ValueError):
    """Error type for when a ledger can't be parsed or has a fatally invalid entry."""


def _write_normalized_entries_to_file(output_path: Path, entries: list[object]) -> None:
    """Writes ledger entries to a file in a normalized format (first sorting the entries)."""
    entries = beancount.core.data.sorted(entries)
    with open(output_path, 'w') as f:
        beancount.parser.printer.print_entries(entries, file=f)


@dataclass
class BeancountLedger:
    """Data structure storing data loaded from a beancount ledger."""

    path: Path  # top-level ledger path
    entries: list[Directive]
    errors: list[object]
    options_map: dict[str, Any]

    @property
    def num_entries(self) -> int:
        """Gets the number of entries."""
        return len(self.entries)

    @staticmethod
    def handle_ledger_errors(errors: list[object]) -> None:
        """Processes a list of errors encountered when parsing a ledger.
        If any is fatal, raises the first such a one as a LedgerError."""
        if errors:
            logger.warning(f'{len(errors)} error(s) found', extra={'highlighter': None})
        for err in errors:
            if _ledger_error_is_fatal(err):
                filename = err.source['filename']  # type: ignore[attr-defined]
                msg = f'error parsing {filename}'
                if hasattr(err, 'message'):
                    msg += f' - {err.message}'
                raise LedgerError(msg)

    @classmethod
    def load_file(cls, ledger_path: str | Path) -> Self:
        """Loads ledger data from a file."""
        path = Path(ledger_path)
        if not has_valid_beancount_extension(path):
            raise ValueError(f'{path} is not a valid beancount file')
        if not path.exists():
            raise ValueError(f'File not found: {path}')
        entries, errors, options_map = loader.load_file(path)
        cls.handle_ledger_errors(errors)
        return cls(path, entries, errors, options_map)

    def get_directives_of_type(self, tp: type[Directive]) -> list[Directive]:
        """Gets a list of all directives of the given type."""
        return [entry for entry in self.entries if isinstance(entry, tp)]

    def num_directives_of_type(self, tp: type[Directive]) -> int:
        """Gets the number of directives of the given type."""
        return sum(1 for entry in self.entries if isinstance(entry, tp))

    @property
    def open_directives(self) -> list[Directive]:
        """Gets a list of all Open directives, which store account information."""
        return self.get_directives_of_type(Open)

    @property
    def num_accounts(self) -> int:
        """Gets the number of accounts in the ledger.
        NOTE: this includes accounts that are closed."""
        return self.num_directives_of_type(Open)

    @property
    def num_transactions(self) -> int:
        """Gets the number of transactions in the ledger."""
        return self.num_directives_of_type(Transaction)

    def apply_file_extension(self, extension: str) -> 'BeancountLedger':
        """Returns a new BeancountLedger where all file paths have the given extension applied."""
        path = self.path.with_suffix(extension)
        entries = []
        for entry in self.entries:
            entry = deepcopy(entry)
            if entry.meta and (filename := entry.meta.get('filename')):
                entry.meta['filename'] = _apply_file_extension(filename, extension)
            entries.append(entry)
        # TODO: change paths in errors too?
        options_map = deepcopy(self.options_map)
        if filename := options_map.get('filename'):
            options_map['filename'] = _apply_file_extension(filename, extension)
        if includes := options_map.get('include'):
            assert isinstance(includes, list)
            options_map['include'] = [_apply_file_extension(filename, extension) for filename in includes]
        return BeancountLedger(path, entries, deepcopy(self.errors), options_map)

    @property
    def non_default_options(self) -> dict[str, Any]:
        """Gets the subset of options_map that does not match the beancount defaults."""
        defaults = options.OPTIONS_DEFAULTS
        return {key: val for (key, val) in self.options_map.items() if (val != defaults.get(key))}

    @property
    def included_files(self) -> list[Path]:
        """Gets the list of files included in the ledger."""
        includes = self.options_map.get('include', [])
        return [path for filename in includes if ((path := Path(filename)) != self.path.resolve())]

    @property
    def _non_default_option_directives(self) -> list[str]:
        """Gets option directives to append to the top-level ledger file (only for options that do not match the beancount defaults)."""
        lines = []
        if options_map := self.non_default_options:
            options_map = {key: val for (key, val) in options_map.items() if (key not in EXEMPT_OPTION_KEYS) and val}
            for key, val in options_map.items():
                line = 'option ' + quote(key)
                if isinstance(val, list):
                    line += ' '.join(map(quote, val))
                else:
                    line += quote(val)
                lines.append(line)
        return lines

    @property
    def _plugin_directives(self) -> list[str]:
        """Gets the list of plugin directives to append to the top-level ledger file."""
        lines = []
        for plugin, config_str in self.options_map.get('plugin', []):
            line = f'plugin "{plugin}"'
            if config_str:
                line += ' ' + quote(config_str)
            lines.append(line)
        return lines

    @property
    def _include_directives(self) -> list[str]:
        """Gets the list of include directives to append to the top-level ledger file."""
        lines = []
        if includes := self.included_files:
            for filename in includes:
                lines.append('include ' + quote(filename))
        return lines

    def save(self, extension: Optional[str] = None, files: Optional[list[Path]] = None) -> None:
        """Saves the ledger contents to one or more files.
        Each entry is saved to its associated filename.
        If an extension is provided, applies this extension to each file.
        If a list of files is provided (which are all included in the ledger), only saves files corresponding to these ones."""
        ledger = self if (extension is None) else self.apply_file_extension(extension)
        if files is None:
            valid_paths = None
        else:
            valid_paths = {path.resolve() for path in files}
            if extension:
                valid_paths = {path.with_suffix(extension) for path in valid_paths}
        is_valid_path: Callable[[Path], bool] = lambda path: (valid_paths is None) or (path.resolve() in valid_paths)
        entries_by_path = defaultdict(list)
        for entry in ledger.entries:
            path = Path(entry.meta['filename'])
            # exclude paths that are not in the specified subset
            if is_valid_path(path):
                entries_by_path[path].append(entry)
        logger.info('Saving ledger file(s)', extra={'highlighter': None})
        for path in sorted(entries_by_path):
            logger.info(f'    {path}')
            _write_normalized_entries_to_file(path, entries_by_path[path])
        if is_valid_path(ledger.path):
            # if there are options/plugins/includes, append them to the top-level ledger file
            with open(ledger.path, 'a') as f:
                for lines in [self._non_default_option_directives, self._plugin_directives, self._include_directives]:
                    if lines:
                        print('', file=f)
                        write_lines(lines, f)

    def apply_transforms(self, transforms: Sequence[LedgerTransform]) -> 'BeancountLedger':
        """Applies a sequence of transforms to a beancount ledger."""
        ledger = self
        if transforms:
            for transform in transforms:
                ledger = transform(ledger)
            if not ledger.entries:
                logger.warning('All entries were filtered out.')
        return ledger



from beancount.core.data import Directive
from beancount.parser import parser
import pytest

from garbanzo.filter import EntryFilter, FixTransaction, RenameTransfers, trim_whitespace


def parse_entry(string: str) -> Directive:
    """Parses a single beancount entry from a string."""
    entries, errors, _ = parser.parse_string(string)
    assert len(entries) == 1
    assert not errors
    return entries[0]


# example filters

PIZZA_FILTER = FixTransaction.from_dict({
    "match": {
        "narration": "PIZZA"
    },
    "set": {
        "account": "Expenses:Restaurant:Pizza"
    }
})

CHARITY_FILTER = FixTransaction.from_dict({
    "match": {
        "narration": "HUMAN\\s*FUND"
    },
    "set": {
         "narration": "Donation",
         "payee": "The Human Fund",
         "account": "Expenses:Charity"
    }
})

STOCK_BUY_FILTER = FixTransaction.from_dict({
    "match": {
        "narration": "BUYMF|BUYSTOCK",
        "account_prefix": "Assets:Brokerage"
    },
    "set": {
        "narration": "Stock Buy"
    }
})

CHAINED_FILTER = EntryFilter.chain([PIZZA_FILTER, CHARITY_FILTER, STOCK_BUY_FILTER])

# example entries

PIZZA_TXN = parse_entry("""
2024-03-14 * "Vincent's Pizza"
  Liabilities:CreditCard:Visa -52.17 USD
  Expenses:Uncategorized       52.17 USD
""")

CHARITY_TXN = parse_entry("""
2024-12-23 * "HUMANFUND ORG"
  Assets:Bank:Checking    -1000.00 USD
  Expenses:Uncategorized   1000.00 USD
""")

STOCK_BUY_TXN = parse_entry("""
2024-12-27 * "BUYSTOCK"
  Assets:Brokerage:Cash     -1233.05 USD
  Assets:Brokerage:RothIRA         1 COKE {1233.05 USD}
""")

PRICE_ENTRY = parse_entry('2024-02-07 balance Assets:Checking  163.00 USD')


@pytest.mark.parametrize(['filt', 'old_entry', 'new_entry'], [
    (
        PIZZA_FILTER,
        PIZZA_TXN,
        parse_entry("""
2024-03-14 * "Vincent's Pizza"
  Liabilities:CreditCard:Visa  -52.17 USD
  Expenses:Restaurant:Pizza     52.17 USD
        """)
    ),
    (
        CHARITY_FILTER,
        CHARITY_TXN,
        parse_entry("""
2024-12-23 * "The Human Fund" "Donation"
  Assets:Bank:Checking  -1000.00 USD
  Expenses:Charity       1000.00 USD
        """)
    ),
    (
        STOCK_BUY_FILTER,
        STOCK_BUY_TXN,
        parse_entry("""
2024-12-27 * "Stock Buy"
  Assets:Brokerage:Cash     -1233.05 USD
  Assets:Brokerage:RothIRA         1 COKE {1233.05 USD}
        """)
    ),
    # no effect
    (
        PIZZA_FILTER,
        CHARITY_TXN,
        CHARITY_TXN
    ),
    # chain of filters
    (
        CHAINED_FILTER,
        CHARITY_TXN,
        parse_entry("""
2024-12-23 * "The Human Fund" "Donation"
  Assets:Bank:Checking  -1000.00 USD
  Expenses:Charity       1000.00 USD
        """)
    ),
    # mismatched types (no effect)
    (
        PIZZA_FILTER,
        PRICE_ENTRY,
        PRICE_ENTRY
    ),
    (
        CHAINED_FILTER,
        PRICE_ENTRY,
        PRICE_ENTRY
    ),
])
def test_fix_transaction(filt, old_entry, new_entry):
    """Tests that the application of a FixTransaction filter (or chain of them) has the expected effect on an entry."""
    assert isinstance(old_entry, type(new_entry))
    assert filt(old_entry) == new_entry


TXN_WITH_EXCESS_WHITESPACE = """
2025-01-01 * " PAYMENT TO   XYZ "
  Assets:Checking  -100 USD
  Assets:Savings    100 USD
"""

@pytest.mark.parametrize(['old_entry', 'new_entry'], [
    (
        parse_entry("""
2025-01-01 * " PAYMENT TO   XYZ "
  Assets:Checking         -100 USD
  Expenses:Uncategorized   100 USD"""),
        parse_entry("""
2025-01-01 * "PAYMENT TO XYZ"
  Assets:Checking         -100 USD
  Expenses:Uncategorized   100 USD"""),
    ),
    # no effect
    (PIZZA_TXN, PIZZA_TXN),
    (PRICE_ENTRY, PRICE_ENTRY),
])
def test_trim_whitespace(old_entry, new_entry):
    assert trim_whitespace(old_entry) == new_entry


ACCOUNT_NAME_MAP = {
    'Assets:Checking': 'Checking',
    'Assets:Savings': 'Savings',
    'Liabilities:CreditCard': 'Credit Card'
}


@pytest.mark.parametrize(['account_name_map', 'old_entry', 'new_entry'], [
    (
        ACCOUNT_NAME_MAP,
        parse_entry("""
2025-01-01 * "Online Transfer to xxxxx"
  Assets:Checking   -250 USD
  Assets:Savings     250 USD
        """),
        parse_entry("""
2025-01-01 * "Transfer: Checking -> Savings"
  Assets:Checking   -250 USD
  Assets:Savings     250 USD
        """)
    ),
    (
        ACCOUNT_NAME_MAP,
        parse_entry("""
2025-01-01 * "Web payment"
  Assets:Checking        -250 USD
  Liabilities:CreditCard  250 USD
        """),
        parse_entry("""
2025-01-01 * "Payment: Checking -> Credit Card"
  Assets:Checking        -250 USD
  Liabilities:CreditCard  250 USD
        """)
    ),
])
def test_rename_transfers(account_name_map, old_entry, new_entry):
    """Tests the RenameTransfers filter."""
    filt = RenameTransfers(account_name_map)
    assert filt(old_entry) == new_entry



from dataclasses import dataclass, field
import datetime
import os
from pathlib import Path
import subprocess
import webbrowser

from beancount.core.data import Account
from fancy_dataclass import CLIDataclass
import rich
from rich.markdown import Markdown
from rich.prompt import InvalidResponse, Prompt, PromptBase
from rich.text import Text

from garbanzo import logger
from garbanzo.helpers import get_relative_account_dir, string_is_date
from garbanzo.ledger import BeancountLedger


class SkipAccountError(Exception):
    """Exception raised to skip an account."""

    pass


class ConfirmWithSkip(PromptBase[bool]):
    """Rich prompt that acts similarly to 'Confirm' but adds a third option to skip
    an account by raising an exception.
    """

    response_type = bool
    validate_error_message = '[prompt.invalid]Please enter Y, N, or S to skip this account'
    choices: list[str] = ['y', 'n', 's']

    def render_default(self, default: bool) -> Text:  # type: ignore[override]
        """Render the default as (y) or (n) rather than True/False."""
        yes, no, _ = self.choices
        return Text(f'({yes})' if default else f'({no})', style='prompt.default')

    def process_response(self, value: str) -> bool:
        """Convert choices to a bool."""
        value = value.strip().lower()
        if value not in self.choices:
            raise InvalidResponse(self.validate_error_message)
        if value == self.choices[-1]:
            raise SkipAccountError()
        return value == self.choices[0]


@dataclass
class AccountMetadata:
    """Account metadata relevant to the download helper."""

    account: Account
    url: str | None = None
    manual_edit: bool = False


def _edit_file(filepath: Path) -> None:
    """Open the given file in the user's default $EDITOR.
    Wait for the editor to close before resuming execution.

    :param filepath: Path to the file to be edited.
    """
    editor = os.environ.get('EDITOR', 'nano')

    try:
        subprocess.run([editor, str(filepath)], check=True)
    except FileNotFoundError:
        logger.error(f'Editor {editor!r} not found')
    except Exception as e:
        logger.error(str(e))


def _add_date_prefix_to_file_name(file_name: str) -> str:
    if len(file_name) > 10 and string_is_date(file_name[:10]):
        return file_name

    today_date = datetime.datetime.now().strftime('%Y-%m-%d')
    return f'{today_date}-{file_name}'


def _move_and_rename_files(files: list[Path], dir_path: Path) -> None:
    for filepath in files:
        new_name = dir_path / _add_date_prefix_to_file_name(filepath.name)
        dir_path.mkdir(parents=True, exist_ok=True)
        filepath.rename(new_name)
        rich.print(f'{filepath.name} -> {new_name.name}')


def _find_recent_files(directory: Path, since_datetime: datetime.datetime) -> list[Path]:
    directory_path = Path(directory)
    recent_files = []

    for file_path in directory_path.iterdir():
        if not file_path.is_file():
            continue
        if file_path.name.startswith('.'):
            continue
        creation_time = datetime.datetime.fromtimestamp(file_path.stat().st_ctime)
        if creation_time > since_datetime:
            recent_files.append(file_path)

    return recent_files


def _handle_account(account: AccountMetadata, download_dir: Path, data_dir: Path, transactions_file: Path | None) -> None:
    if not account.url:
        return

    rich.print(f"[bold cyan]Handling:[/] [bold blue]{account.account}[/] [grey](Press 's' at any point to skip)[/]")
    resp = ConfirmWithSkip.ask(f'[bold cyan]Open:[/] [bold]{account.url}[/]', default=True)
    if resp:
        webbrowser.open(account.url)

    cur_time = datetime.datetime.now()
    new_downloads = []
    if not account.manual_edit:
        Prompt.ask('[bold cyan]Waiting for download...[/] (press Enter when file(s) downloaded)')
        new_downloads = _find_recent_files(directory=download_dir, since_datetime=cur_time)
    if account.manual_edit or not new_downloads:
        prompt_str = 'Account marked as manual entry, go to transactions file now?'
        if not account.manual_edit:
            prompt_str = '[yellow]Found no downloaded files.[/] Edit transactions manually instead?'
        resp = ConfirmWithSkip.ask(
            prompt_str,
            default=account.manual_edit,
        )
        if resp:
            if not transactions_file:
                rich.print('[red]No transactions file given[/]')
            else:
                _edit_file(transactions_file)
        return

    files = '\n'.join([f'- {f.name}' for f in new_downloads])
    rich.print('[bold cyan]Found downloaded files:[/]')
    rich.print(Markdown(files))

    directory = data_dir / get_relative_account_dir(account.account)
    resp = ConfirmWithSkip.ask(f'Move file(s) to [bold green]{directory}[/]?', default=True)
    if resp:
        _move_and_rename_files(files=new_downloads, dir_path=directory)

    rich.print('[bold magenta]----------[/]')


def _load_accounts_from_file(accounts_file: Path) -> list[AccountMetadata]:
    ledger = BeancountLedger.load_file(accounts_file)
    metas = []
    for entry in ledger.open_directives:
        is_manual_edit = entry.meta.get('manual_edit', 'false').lower().strip() == 'true'
        metas.append(AccountMetadata(account=entry.account, url=entry.meta.get('url'), manual_edit=is_manual_edit))
    return metas


def download_assistant(accounts_file: Path, data_dir: Path, download_dir: Path | None = None, transactions_file: Path | None = None) -> None:
    """Entrypoint for the download assistant. Given a set of accounts with a
    'url' metadata tag, will guide the user through downloading data for each
    account. When data is downloaded, automatically tags it with a date (so that
    it is picked up by the fava "documents" tab) and moves it into the proper
    directory.
    """
    if not data_dir.exists():
        raise RuntimeError(f'Data directory not found: {data_dir!r}')

    if not download_dir:
        download_dir = Path.home() / 'Downloads'

    for account in _load_accounts_from_file(accounts_file=accounts_file):
        try:
            _handle_account(account=account, data_dir=data_dir, download_dir=download_dir, transactions_file=transactions_file)
        except SkipAccountError:
            rich.print(f'\n[bold yellow]Skipping account:[/] {account.account}\n')
            continue


@dataclass
class DownloadAssistantCmd(CLIDataclass, command_name='dl-assistant'):
    """Download assistant for beancount accounts."""

    accounts_file: Path = field(metadata={'help': 'input file containing accounts'})
    data_dir: Path = field(metadata={'help': 'path to directory where downloads should be moved'})
    download_dir: Path | None = field(default=None, metadata={'help': 'directory where downloads are looked for'})
    transactions_file: Path | None = field(default=None, metadata={'help': 'file to which transactions are written'})

    def run(self) -> None:
        download_assistant(accounts_file=self.accounts_file, data_dir=self.data_dir, download_dir=self.download_dir, transactions_file=self.transactions_file)



from pathlib import Path
import shutil

import pytest

from garbanzo.cli.main import GarbanzoCLI
from garbanzo.ledger import EXEMPT_OPTION_KEYS, BeancountLedger

from . import EXAMPLE_DIR


TRANSACTION_PATH = EXAMPLE_DIR / 'transactions.beancount'
FILTERED_TRANSACTION_PATH = EXAMPLE_DIR / 'transactions.filtered.beancount'
NORMALIZED_TRANSACTION_PATH = EXAMPLE_DIR / 'transactions.normalized.beancount'
FILTER_JSON_PATH = EXAMPLE_DIR / 'filters.json'
TRANSFER_PATH = EXAMPLE_DIR / 'transfers.beancount'
FILTERED_TRANSFER_PATH = EXAMPLE_DIR / 'transfers.filtered.beancount'
LEDGER_PATH = EXAMPLE_DIR / 'ledger.beancount'


def make_local_copy(tmpdir, path):
    """Makes a local copy of the given path, relative to a temporary directory.
    Returns the local path."""
    local_path = Path(tmpdir / path.name)
    shutil.copy(path, local_path)
    return local_path


@pytest.mark.parametrize(['ledger_path', 'filtered_path', 'extra_args', 'error'], [
    (TRANSACTION_PATH, FILTERED_TRANSACTION_PATH, ['--json-filters', str(FILTER_JSON_PATH)], None),
    (TRANSACTION_PATH, FILTERED_TRANSACTION_PATH, ['--filters', '--json-filters', str(FILTER_JSON_PATH)], None),
    (TRANSFER_PATH, FILTERED_TRANSFER_PATH, [], None),
    (TRANSFER_PATH, None, ['--filters'], 'No filters provided'),
])
def test_filter(tmpdir, caplog, ledger_path, filtered_path, extra_args, error):
    """Tests the filter subcommand."""
    # make temporary copy of ledger file
    ledger_path_copy = make_local_copy(tmpdir, ledger_path)
    args = ['filter', str(ledger_path_copy)] + extra_args
    if error:  # expect error
        with pytest.raises(SystemExit):
            GarbanzoCLI.main(args)
        assert error in caplog.text
    else:
        filt_str = filtered_path.read_text()
        GarbanzoCLI.main(args)
        assert filt_str == ledger_path_copy.with_suffix('.filt.beancount').read_text()
        args = ['filter', str(ledger_path_copy), '--inplace'] + extra_args
        GarbanzoCLI.main(args)
        assert filt_str == ledger_path_copy.read_text()


def test_normalize_invalid_file(caplog):
    """Tests the normalize subcommand on a file that doesn't exist."""
    args = ['normalize', 'fake_ledger.beancount']
    try:
        GarbanzoCLI.main(args)
    except SystemExit as e:
        assert e.code == 1  # noqa: PT017
        assert 'File not found: fake_ledger.beancount' in caplog.text
    else:  # pragma: no cover
        pytest.fail('expected error')


def test_normalize_single(tmpdir):
    """Tests the normalize subcommand on a single transaction file."""
    norm_str = NORMALIZED_TRANSACTION_PATH.read_text()
    transaction_path = make_local_copy(tmpdir, TRANSACTION_PATH)
    args = ['normalize', str(transaction_path)]
    GarbanzoCLI.main(args)
    assert norm_str == transaction_path.with_suffix('.norm.beancount').read_text()
    args = ['normalize', str(transaction_path), '--inplace']
    GarbanzoCLI.main(args)
    assert norm_str == transaction_path.read_text()


def test_normalize_multi(tmpdir):
    """Tests the normalize subcommand on a top-level ledger file with options/plugins/includes."""
    ledger_path = make_local_copy(tmpdir, LEDGER_PATH)
    transaction_path = make_local_copy(tmpdir, TRANSACTION_PATH)
    ledger = BeancountLedger.load_file(ledger_path)
    assert not ledger.errors
    args = ['normalize', str(ledger_path)]
    GarbanzoCLI.main(args)
    norm_ledger_path = ledger_path.with_suffix('.norm.beancount')
    assert norm_ledger_path.exists()
    assert transaction_path.with_suffix('.norm.beancount').exists()
    norm_ledger = BeancountLedger.load_file(norm_ledger_path)
    assert not norm_ledger.errors
    assert len(ledger.entries) == len(norm_ledger.entries)
    # test entries are equivalent up to filenames
    def _fix_entry(entry):
        if hasattr(entry, 'postings'):
            postings = [posting._replace(meta=None) for posting in entry.postings]
            entry = entry._replace(postings=postings)
        return entry._replace(meta=None)
    assert list(map(_fix_entry, ledger.entries)) == list(map(_fix_entry, norm_ledger.entries))
    # test options_maps are equal (for non-exempt keys)
    opts1 = {key: val for (key, val) in ledger.options_map.items() if (key not in EXEMPT_OPTION_KEYS)}
    opts2 = {key: val for (key, val) in norm_ledger.options_map.items() if (key not in EXEMPT_OPTION_KEYS)}
    assert opts1 == opts2
    # test plugins are equal
    assert ledger.options_map['plugin'] == norm_ledger.options_map['plugin'] == [('beancount.plugins.auto_accounts', None)]


def test_normalize_subset(tmpdir, caplog):
    """Tests the normalize subcommand on a top-level ledger file with options/includes, but only modify an included file."""
    ledger_path = make_local_copy(tmpdir, LEDGER_PATH)
    transaction_path = make_local_copy(tmpdir, TRANSACTION_PATH)
    ledger = BeancountLedger.load_file(ledger_path)
    assert not ledger.errors
    args = ['normalize', str(ledger_path), '--files', str(transaction_path)]
    GarbanzoCLI.main(args)
    # top-level file was not normalized
    assert not ledger_path.with_suffix('.norm.beancount').exists()
    # transaction file was normalized
    norm_transaction_path = transaction_path.with_suffix('.norm.beancount')
    assert NORMALIZED_TRANSACTION_PATH.read_text() == norm_transaction_path.read_text()
    # include invalid file
    args = ['normalize', str(ledger_path), '--files', str(transaction_path), 'fake_file.beancount']
    with pytest.raises(SystemExit):
        GarbanzoCLI.main(args)
    assert 'File not found in ledger: fake_file.beancount' in caplog.text



from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path
from typing import Optional

from fancy_dataclass import CLIDataclass

from garbanzo import logger
from garbanzo.ledger import BeancountLedger


@dataclass
class ModifyLedgerCmd(CLIDataclass):
    """Base class for a subcommand that reads in a beancount ledger and modifies it in some way."""

    ledger_file: Path = field(metadata={'help': 'beancount ledger file'})
    inplace: bool = field(default=False, metadata={'help': 'modify files in place'})
    files: Optional[list[Path]] = field(default=None, metadata={'nargs': '+', 'help': 'subset of files to modify', 'metavar': 'FILE'})

    def __post_init__(self) -> None:
        if self.files is not None:
            self._check_files(self.ledger)

    def _check_files(self, ledger: BeancountLedger) -> None:
        """Checks that the list of files provided are among those included in the ledger.
        If not, raises a ValueError."""
        assert self.files is not None
        valid_paths = {path.resolve() for path in [ledger.path] + ledger.included_files}
        for path in self.files:
            assert isinstance(path, Path)
            if path.resolve() not in valid_paths:
                raise ValueError(f'File not found in ledger: {path}')

    @cached_property
    def ledger(self) -> BeancountLedger:
        return BeancountLedger.load_file(self.ledger_file)


@dataclass
class NormalizeCmd(ModifyLedgerCmd, command_name='normalize'):
    """Sort entries, normalize whitespace and currency alignment."""

    def run(self) -> None:
        logger.info(f'Normalizing ledger {self.ledger_file}')
        extension = None if self.inplace else '.norm.beancount'
        self.ledger.save(extension=extension, files=self.files)



from collections.abc import Iterable
import datetime
from pathlib import Path
from typing import IO, Any, Hashable, TypeVar


K = TypeVar('K', bound=Hashable)


def replace_dict_keys(d: dict[K, Any], key_map: dict[K, K]) -> dict[K, Any]:
    """Replaces keys in a dict with new ones, according to the given key_map.
    NOTE: it is required that key_map is one-to-one."""
    vals = list(key_map.values())
    if len(set(vals)) != len(vals):
        raise ValueError('key_map must be one-to-one')
    return {key_map.get(key, key): val for (key, val) in d.items()}


def get_relative_account_dir(account: str) -> Path:
    """Gets the (relative) subdirectory of the documents directory corresponding to the given account."""
    path = Path()
    for seg in account.strip().split(':'):
        path = path / seg
    return path


def string_is_date(string: str, date_format: str = '%Y-%m-%d') -> bool:
    """Checks if a string can be converted to a date."""
    try:
        _ = datetime.datetime.strptime(string, date_format)
        return True
    except ValueError:
        return False


def quote(val: Any) -> str:
    """Renders a value as a string in double quotes."""
    return f'"{val}"'


def write_lines(lines: Iterable[str], f: IO[str]) -> None:
    """Writes lines to a file."""
    for line in lines:
        print(line, file=f)



import logging

from rich.console import Console
from rich.logging import RichHandler


__version__ = '0.1.0'

PROG = 'garbanzo'

LOG_FMT = '%(message)s'

# create global logger

console = Console(stderr=True)

handler = RichHandler(
    show_time=False,
    show_level=True,
    show_path=False,
    markup=True,
    console=console,
)

logging.basicConfig(
    level=logging.INFO,
    format=LOG_FMT,
    handlers=[handler],
)

logger = logging.getLogger(PROG)



from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

from beancount.core.data import Directive

from garbanzo import logger
from garbanzo.cli.normalize import ModifyLedgerCmd
from garbanzo.filter import EntryFilter, filter_transform, load_json_filters, rename_transfers, trim_whitespace
from garbanzo.ledger import LedgerTransform


# map from filter names to filters (as LedgerTransforms)
FILTERS: dict[str, LedgerTransform] = {
    'rename_transfers': rename_transfers,
    'trim_whitespace': filter_transform(trim_whitespace),
}

# map from filter names to descriptions
FILTER_DESCRS: dict[str, str] = {
    'rename_transfers': 'use consistent format for narration of "transfer" transactions, based on "name" account metadata',
    'trim_whitespace': 'remove excess whitespace from narration and payee',
}

# list of filters to use by default
DEFAULT_FILTERS = [
    'rename_transfers',
    'trim_whitespace',
]

_FILTER_DESCR = '\n'.join(f'    {name}: {descr}' for (name, descr) in FILTER_DESCRS.items())
_FILTER_DESCR += '\n(default: ' + ', '.join(DEFAULT_FILTERS) + ')'


@dataclass
class FilterCmd(ModifyLedgerCmd, command_name='filter'):
    """Apply one or more filters to remove or modify ledger entries."""

    filters: list[str] = field(default_factory=lambda: DEFAULT_FILTERS, metadata={'nargs': '*', 'choices': list(FILTERS), 'help': f'filters to apply\n{_FILTER_DESCR}'})
    # TODO: store JSON paths within ledger metadata to avoid using a special command-line argument here
    json_filters: list[Path] = field(default_factory=list, metadata={'nargs': '+', 'help': 'one or more JSON files containing a list of filters'})

    def get_json_filter(self) -> Optional[EntryFilter[Directive]]:
        """Gets a single EntryFilter object used to filter ledger entries."""
        filters: list[EntryFilter[Directive]] = []
        for json_path in self.json_filters:
            filters.extend(load_json_filters(json_path))
        return EntryFilter.chain(filters) if filters else None

    def run(self) -> None:
        logger.info(f'Filtering ledger {self.ledger_file}')
        if self.filters:
            filter_str = ', '.join(self.filters)
            logger.info(f'Using filters: {filter_str}')
        if self.json_filters:
            json_strs = ', '.join(map(str, self.json_filters))
            logger.info(f'Using filters from {json_strs}')
        transforms = [FILTERS[name] for name in self.filters]
        json_filter = self.get_json_filter()
        if json_filter is not None:
            transforms.append(filter_transform(json_filter))
        if not transforms:
            raise ValueError('No filters provided.')
        logger.info(f'Pre-filter:  {self.ledger.num_entries} entries, {self.ledger.num_transactions} transactions')
        ledger = self.ledger.apply_transforms(transforms)
        logger.info(f'Post-filter: {ledger.num_entries} entries, {ledger.num_transactions} transactions')
        extension = None if self.inplace else '.filt.beancount'
        ledger.save(extension=extension, files=self.files)



#!/usr/bin/env python3

from argparse import RawTextHelpFormatter
from dataclasses import dataclass, field
import sys
from typing import Optional, Union

from fancy_dataclass import CLIDataclass

from garbanzo import __version__, logger
from garbanzo.cli.download_assistant import DownloadAssistantCmd
from garbanzo.cli.filter import FilterCmd
from garbanzo.cli.normalize import NormalizeCmd


@dataclass
class GarbanzoCLI(CLIDataclass, version=f'%(prog)s {__version__}', formatter_class=RawTextHelpFormatter):
    """Assorted utilities for interacting with a beancount ledger."""

    subcommand: Union[
        DownloadAssistantCmd,
        FilterCmd,
        NormalizeCmd,
    ] = field(metadata={'subcommand': True})

    @classmethod
    def main(cls, arg_list: Optional[list[str]] = None) -> None:
        """Add custom error handling to main function to exit gracefully when possible."""
        try:
            super().main(arg_list=arg_list)  # delegate to subcommand
        except Exception as e:
            logger.error(str(e))
            sys.exit(1)


if __name__ == '__main__':
    GarbanzoCLI.main()
